import asyncio
from typing import Callable, TypeVar, Optional

from adhesive.model.AdhesiveProcess import AdhesiveProcess
from adhesive.consoleui.ConsoleUserTaskProvider import ConsoleUserTaskProvider
from adhesive.model.WorkflowExecutor import WorkflowExecutor
from adhesive.steps.AdhesiveTask import AdhesiveTask
from adhesive.steps.AdhesiveUserTask import AdhesiveUserTask
from adhesive.process_read.bpmn import read_bpmn_file
from adhesive.process_read.tasks import generate_from_tasks

T = TypeVar('T')
process = AdhesiveProcess('_root')


def task(*task_names: str) -> Callable[..., Callable[..., T]]:
    def wrapper_builder(f: Callable[..., T]) -> Callable[..., T]:
        process.steps.append(AdhesiveTask(f, *task_names))
        return f

    return wrapper_builder


def usertask(*task_names: str) -> Callable[..., Callable[..., T]]:
    def wrapper_builder(f: Callable[..., T]) -> Callable[..., T]:
        process.steps.append(AdhesiveUserTask(f, *task_names))
        return f

    return wrapper_builder


def build(ut_provider: Optional['UserTaskProvider'] = None,
          wait_tasks: bool=True):
    process.workflow = generate_from_tasks(process)

    return _build(ut_provider=ut_provider,
                  wait_tasks=wait_tasks)


def bpmn_build(file_name: str,
               ut_provider: Optional['UserTaskProvider'] = None,
               wait_tasks: bool=True):
    """ Start a build that was described in BPMN """
    process.workflow = read_bpmn_file(file_name)

    return _build(ut_provider=ut_provider,
                  wait_tasks=wait_tasks)


def _build(ut_provider: Optional['UserTaskProvider'] = None,
           wait_tasks: bool=True):

    if ut_provider is None:
        ut_provider = ConsoleUserTaskProvider()

    fn = WorkflowExecutor(
        process,
        ut_provider=ut_provider,
        wait_tasks=wait_tasks).execute()

    return asyncio.get_event_loop().run_until_complete(fn)


def bpmn_build_async(
        file_name: str,
        ut_provider: Optional['UserTaskProvider'] = None,
        wait_tasks: bool=True):
    """ Start a build that was described in BPMN """
    process.workflow = read_bpmn_file(file_name)

    if ut_provider is None:
        ut_provider = ConsoleUserTaskProvider()

    return WorkflowExecutor(
        process,
        ut_provider=ut_provider,
        wait_tasks=wait_tasks).execute()


from adhesive.model.UserTaskProvider import UserTaskProvider



import sys
import unittest
import uuid

from adhesive.logredirect.LogRedirect import redirect_stdout
from adhesive.storage.ensure_folder import ensure_folder


class TestLogRedirection(unittest.TestCase):
    def test_file_redirection_logs(self):
        original_stdout = sys.stdout
        original_stderr = sys.stderr

        target_folder = ensure_folder(str(uuid.uuid4()))

        self.assertEqual(original_stdout, sys.stdout)
        self.assertEqual(original_stderr, sys.stderr)
        print("x")

        with redirect_stdout(target_folder):
            self.assertNotEqual(original_stdout, sys.stdout)
            self.assertNotEqual(original_stderr, sys.stderr)
            print("y")

        self.assertEqual(original_stdout, sys.stdout)
        self.assertEqual(original_stderr, sys.stderr)
        print("z")


if __name__ == '__main__':
    unittest.main()



from typing import cast

from adhesive.graph.ScriptTask import ScriptTask
from adhesive.logredirect.LogRedirect import redirect_stdout
from adhesive.model.ActiveEvent import ActiveEvent
from adhesive.steps.ExecutionToken import ExecutionToken


def call_script_task(event: ActiveEvent) -> ExecutionToken:
    with redirect_stdout(event):
        exec(
            cast(ScriptTask, event.task).script,
            {},                          # globals
            {
                 "context": event.context,
                 "loop": event.context.loop,
                 "data": event.context.data,
            })  # locals

        return event.context



import adhesive
from adhesive.config import LocalConfigReader

_globals = globals()
_locals = locals()

# read local configuration.
adhesive.config.current =\
    LocalConfigReader.read_configuration()

def main():
    with open('_adhesive.py', 'r', encoding='utf-8') as f:
        content = f.read()

    exec(content, _globals, _locals)


if __name__ == "__main__":
    main()



import unittest

from adhesive.model.WorkflowExecutor import WorkflowExecutor
from adhesive.process_read.bpmn import read_bpmn_file

from test.adhesive.steps.test_tasks import adhesive, _async
from test.adhesive.steps.check_equals import assert_equal_steps


class TestTaskJoinExecution(unittest.TestCase):
    """
    Test if the workflow executor can process exclusive gateways.
    """
    def test_task_join_execution(self):
        """
        Load a workflow with a gateway and test it..
        """
        adhesive.process.workflow = read_bpmn_file("test/adhesive/xml/task-join.bpmn")

        workflow_executor = WorkflowExecutor(adhesive.process)
        data = _async(workflow_executor.execute())

        assert_equal_steps({
            'Build Germanium Image': 1,
            'Test Chrome': 1,
            'Test Firefox': 3,
        }, data.steps)

        self.assertFalse(workflow_executor.events,
                         "Some events were not unregistered and/or executed.")

    def test_task_join_execution_non_wait(self):
        """
        Load a workflow with a gateway and test it..
        """
        adhesive.process.workflow = read_bpmn_file("test/adhesive/xml/task-join.bpmn")

        workflow_executor = WorkflowExecutor(adhesive.process, wait_tasks=False)
        data = _async(workflow_executor.execute())

        assert_equal_steps({
            'Build Germanium Image': 3,  # 1 chrome + 2 firefox
            'Test Chrome': 1,
            'Test Firefox': 3,
        }, data.steps)

        self.assertFalse(workflow_executor.events,
                         "Some events were not unregistered and/or executed.")


if __name__ == '__main__':
    unittest.main()



import unittest

from adhesive.config.LocalConfigReader import read_configuration


class TestConfigLoading(unittest.TestCase):
    def test_local_config_reading(self):
        config = read_configuration(
            cwd="test/adhesive/config/tc_local/local",
            environment={
                "HOME": "test/adhesive/config/tc_local/user",
                "ADHESIVE_ENVIRONMENT": "environment",
                "ADHESIVE_ENVIRONMENT_LIST": "environment_environment_list_a:environment_environment_list_b"
            }
        )

        self.assertEqual("local", config.local)
        self.assertEqual("user", config.user)
        self.assertEqual("environment", config.environment)
        self.assertEqual("/tmp", config.temp_folder)
        self.assertEqual([
            "environment_environment_list_a",
            "environment_environment_list_b",
        ], config.environment_list)
        self.assertEqual([
            "local_local_list_a",
            "local_local_list_b",
        ], config.local_list)
        self.assertEqual([
            "user_user_list_a",
            "user_user_list_b",
        ], config.user_list)

        self.assertIsNone(config.not_existing)



import re
from typing import Callable, Optional, List

from adhesive.graph.BaseTask import BaseTask


class AdhesiveBaseTask:
    """
    A task implementation.
    """
    def __init__(self,
                 code: Callable,
                 *expressions: str) -> None:
        self.re_expressions = list(map(re.compile, expressions))
        self.code = code
        self.expressions = expressions

    def matches(self,
                task: BaseTask,
                resolved_name: str) -> Optional[List[str]]:
        """
        Checks if this implementation matches any of the expressions
        bounded to this task. If yes, it returns the potential variables
        extracted from the expression.
        :param context:
        :return:
        """
        for re_expression in self.re_expressions:
            m = re_expression.match(resolved_name)

            if m:
                return list(m.groups())

        return None



from typing import Dict, Any
import os

from adhesive.config.LocalAdhesiveConfig import LocalAdhesiveConfig
import yaml


def read_configuration(cwd=".",
                       environment=os.environ) -> LocalAdhesiveConfig:
    user_config_path = os.path.join(
        environment.get("HOME", ""),
        ".adhesive", "config.yml"
    )
    user_config = read_configuration_from_file(user_config_path)

    local_config_path = os.path.join(cwd, ".adhesive", "config.yml")
    local_config = read_configuration_from_file(local_config_path)

    return LocalAdhesiveConfig(
        user_config=user_config,
        local_config=local_config,
        environment=environment,
    )


def read_configuration_from_file(path: str) -> Dict[str, Any]:
    try:
        with open(path, "rt") as f:
            return yaml.safe_load(f)
    except Exception as e:
        print(e)
        return dict()



from adhesive import AdhesiveTask, AdhesiveUserTask
from adhesive.graph.Edge import Edge
from adhesive.graph.EndEvent import EndEvent
from adhesive.graph.StartEvent import StartEvent
from adhesive.graph.Task import Task
from adhesive.graph.UserTask import UserTask
from adhesive.graph.Workflow import Workflow
from adhesive.model.AdhesiveProcess import AdhesiveProcess


current_id = 0


def next_id():
    global current_id

    current_id += 1
    return f"_{current_id}"


def generate_from_tasks(process: AdhesiveProcess) -> Workflow:
    workflow = Workflow(next_id())

    if not process.steps:
        raise Exception("No task was defined. You need to create "
                        "tasks with @adhesive.task or @adhesive.usertask .")

    last_task = StartEvent(next_id(), "start event")
    workflow.add_start_event(last_task)

    for step in process.steps:
        if isinstance(step, AdhesiveTask):
            task = Task(next_id(), step.expressions[0])
            workflow.add_task(task)
        elif isinstance(step, AdhesiveUserTask):
            task = UserTask(next_id(), step.expressions[0])
            workflow.add_task(task)
        else:
            raise Exception(f"Unsupported task {step}")

        edge = Edge(next_id(), last_task.id, task.id)
        workflow.add_edge(edge)
        last_task = task

    task = EndEvent(next_id(), "end event")
    workflow.add_end_event(task)

    edge = Edge(next_id(), last_task.id, task.id)
    workflow.add_edge(edge)

    return workflow



from typing import Any

from adhesive.config.AdhesiveConfig import AdhesiveConfig


class AdhesiveConfigResolver:
    """
    This only holds the wrapper over getattr to allow
    fetching settings directly.
    """
    def __init__(self, adhesive_config: AdhesiveConfig):
        self.config = adhesive_config

    def __getattr__(self, item: str) -> Any:
        return self.config.get_attribute(item)



import os
from typing import Union

from adhesive import config


def ensure_folder(item: Union['Workspace', 'ActiveEvent', str]) -> str:
    """
    Ensures the folder for the given item exists.

    :param item:
    :return:
    """
    full_path = get_folder(item)
    os.makedirs(full_path, exist_ok=True)

    return full_path


def get_folder(item: Union['Workspace', 'ActiveEvent', str]) -> str:
    if isinstance(item, Workspace):
        return os.path.join(
            config.current.temp_folder,
            item.execution.id,
            "workspaces",
            item.id)

    if isinstance(item, ActiveEvent):
        # FIXME: loop/parent loop check?
        return os.path.join(
            config.current.temp_folder,
            item.context.execution.id,
            "logs",
            item.task.id,
            item.id)

    if isinstance(item, str):
        return os.path.join(config.current.temp_folder, item)

    raise Exception(f"Unable to get_folder for {item}.")


from adhesive.model.ActiveEvent import ActiveEvent
from adhesive.workspace.Workspace import Workspace


import abc
import os
from typing import Any, List


class AdhesiveConfig:
    def __init__(self):
        pass

    @abc.abstractmethod
    def __getattr__(self, item) -> Any:
        pass

    @abc.abstractmethod
    def secret_locations(self) -> List[str]:
        return [
            os.path.join(".", ".adhesive", "secrets"),
            os.path.join(os.environ.get("HOME", "."), ".adhesive", "secrets"),
        ]



import glob
import os
import unittest

from adhesive.model.WorkflowExecutor import WorkflowExecutor
from adhesive.process_read.bpmn import read_bpmn_file
from test.adhesive.steps.check_equals import assert_equal_steps
from test.adhesive.steps.test_tasks import adhesive, _async


class TestIfLogRedirectionWorks(unittest.TestCase):
    """
    Test if the workflow executor can process inclusive gateways.
    """
    def test_log_redirection(self):
        """
        Load a workflow with a gateway and test it..
        """
        adhesive.process.workflow = read_bpmn_file("test/adhesive/xml/redirect-logs.bpmn")

        workflow_executor = WorkflowExecutor(adhesive.process)
        data = _async(workflow_executor.execute())

        assert_equal_steps({
            "sh: echo hello world && echo bad world >&2 && echo good world": 1,
            "Store current execution id": 1,
        }, data.steps)
        self.assertFalse(workflow_executor.events)

        adhesive_temp_folder = os.environ.get("ADHESIVE_TEMP_FOLDER", "/tmp/adhesive")
        log_path = glob.glob(os.path.join(
            adhesive_temp_folder,
            data.execution_id,
            "logs",
            "_4",
            "*",
            "stdout"))

        with open(log_path[0], "rt") as f:
            self.assertEqual(f.read(), "sh: echo hello world && "
                                       "echo bad world >&2 && "
                                       "echo good world\nhello world\ngood world\n")

        log_path = glob.glob(os.path.join(
            adhesive_temp_folder,
            data.execution_id,
            "logs",
            "_4",
            "*",
            "stderr"))

        with open(log_path[0], "rt") as f:
            self.assertEqual(f.read(), "bad world\n")



from adhesive.config import LocalConfigReader
from adhesive.config.LocalAdhesiveConfig import LocalAdhesiveConfig
from adhesive.config.NoopAdhesiveConfig import NoopAdhesiveConfig

current = LocalConfigReader.read_configuration()



import os
from typing import Any, Dict

from adhesive.config.AdhesiveConfig import AdhesiveConfig


class LocalAdhesiveConfig(AdhesiveConfig):
    """
    A configuration that's assembled from the local files.
    """
    def __init__(self,
                 user_config: Dict[str, Any],
                 local_config: Dict[str, Any],
                 environment: Dict[str, str]):
        super(LocalAdhesiveConfig, self).__init__()

        # FIXME: detect if possible where's the /tmp folder
        self._defaults = {
            "temp_folder": "/tmp"
        }
        self._local_config = local_config if local_config is not None else dict()
        self._user_config = user_config if user_config is not None else dict()
        self._environment = environment if environment is not None else os.environ

    def __getattr__(self, item: str) -> Any:
        env_name = f"ADHESIVE_{item.upper()}"

        if env_name in self._environment:
            if env_name.endswith("_LIST"):
                return self._environment[env_name].split(os.path.pathsep)
            return self._environment[env_name]

        if item in self._local_config:
            return self._local_config[item]

        if item in self._user_config:
            return self._user_config[item]

        if item in self._defaults:
            return self._defaults[item]

        return None




import unittest

from adhesive.steps.ExecutionData import ExecutionData


class TestExecutionData(unittest.TestCase):
    def test_workflow_data_merge(self):
        data1 = ExecutionData({
            "list": [1, 2, 3],
            "set": {"a", "b"},
            "dict": {
                "a": "a",
                "b": "b",
            },
            "number": 3,
            "to_remove": 1,
        })

        data2 = ExecutionData({
            "list": [4, 5, 6],
            "set": {"b", "c"},
            "dict": {
                "b": "b2",
                "c": "c",
            },
            "number": 1,
            "to_remove": None
        })

        result = ExecutionData.merge(data1, data2)

        self.assertEqual({
            "list": [4, 5, 6],  # same type, gets overwritten
            "set": {"a", "b", "c"},
            "dict": {
                "a": "a",
                "b": "b2", # keys are taken first from the 2nd object
                "c": "c",
            },
            "number": 1
        }, result.as_dict())



import os
import shlex
import subprocess
import sys
from contextlib import contextmanager
from typing import Optional, Union, Iterable
from uuid import uuid4

from adhesive.storage.ensure_folder import ensure_folder
from .Workspace import Workspace


class DockerWorkspace(Workspace):
    def __init__(self,
                 workspace: Workspace,
                 image_name: str) -> None:
        super(DockerWorkspace, self).__init__(
            execution=workspace.execution,
            pwd=workspace.pwd)

        pwd = workspace.pwd

        self.container_id = subprocess.check_output([
            "docker", "run",
            "-t",
            "-v", f"{pwd}:{pwd}",
            "-d",
            "--entrypoint", "cat",
            "-u", "1000:1000",  # FIXME
            image_name
        ]).decode('utf-8').strip()

    def run(self, command: str) -> None:
        subprocess.check_call(
            [
                "docker", "exec",
                          "-w", self.pwd,
                          self.container_id,
                          "/bin/sh", "-c",
                          command
            ],
            cwd=self.pwd,
            stdout=sys.stdout,
            stderr=sys.stderr)

    def write_file(
            self,
            file_name: str,
            content: str) -> None:
        """
        Write a file on the remote docker instance. Since we can't
        really just write files, we create a temp file, then we
        copy it remotely.
        :param file_name:
        :param content:
        :return:
        """
        try:
            tmp_folder = ensure_folder("tmp")
            tmp_file = os.path.join(tmp_folder, str(uuid4()))
            with open(tmp_file, "wt") as f:
                f.write(content)
            self.copy_to_agent(tmp_file, file_name)
        finally:
            os.remove(tmp_file)

    def rm(self, path: Optional[str]=None) -> None:
        """
        Remove a path from the container. We're calling `rm` to do
        the actual operation.
        :param path:
        :return:
        """
        if not path:
            raise Exception("You need to pass a subpath for deletion")

        self.run(f"rm -fr {shlex.quote(path)}")

    def mkdir(self, path: str = None) -> None:
        full_path = os.path.join(self.pwd, path)
        self.run(f"mkdir -p {shlex.quote(full_path)}")

    def copy_to_agent(self,
                      from_path: str,
                      to_path: str):
        subprocess.check_call(
            [
                "docker", "cp", from_path, f"{self.container_id}:{to_path}"
            ],
            stdout=sys.stdout,
            stderr=sys.stderr)

    def copy_from_agent(self,
                        from_path: str,
                        to_path: str):
        subprocess.check_call(
            [
                "docker", "cp", f"{self.container_id}:{from_path}", to_path
            ],
            stdout=sys.stdout,
            stderr=sys.stderr)

    def _destroy(self):
        subprocess.check_call(
            [
                "docker", "rm", "-f", self.container_id
            ],
            stdout=sys.stdout,
            stderr=sys.stderr)


@contextmanager
def inside(workspace: Workspace,
           image_name: str):
    w = None

    try:
        w = DockerWorkspace(workspace=workspace,
                            image_name=image_name)
        yield w
    finally:
        if w is not None:
            w._destroy()


@contextmanager
def build(workspace: Workspace,
          tags: Union[str, Iterable[str]]):

    # we always consider a list of tags
    if isinstance(tags, str):
        tags = [tags]

    command = "docker build "

    for tag in tags:
        command += f"-t {tag} "

    command += "."

    workspace.run(command)



import os
import sys
from contextlib import contextmanager
from typing import Union, Any

from adhesive.model.ActiveEvent import ActiveEvent


class StreamLogger:
    def __init__(self,
                 name: str,
                 folder: str) -> None:
        if not folder:
            raise Exception(f"No folder specified for output: {folder}")

        self.log = open(
            os.path.join(folder, name),
            "at")

        self._closed = False

    @staticmethod
    def from_event(event: Union[ActiveEvent, str],
                   name: str) -> 'StreamLogger':
        folder = ensure_folder(event)
        return StreamLogger(name, folder)

    @property
    def fileno(self):
        return self.log.fileno

    def flush(self):
        self.log.flush()

    def write(self, message):
        if self._closed:
            raise Exception("already closed")

        self.log.write(message)
        self.log.flush()

    def close(self) -> None:
        self.log.close()
        self._closed = True


@contextmanager
def redirect_stdout(event: Union[ActiveEvent, str]) -> Any:
    old_stdout = sys.stdout
    old_stderr = sys.stderr

    try:
        new_stdout = StreamLogger.from_event(event, "stdout")
        new_stderr = StreamLogger.from_event(event, "stderr")

        sys.stdout = new_stdout
        sys.stderr = new_stderr

        yield None
    finally:
        sys.stdout = old_stdout
        sys.stderr = old_stderr

        new_stdout.close()
        new_stderr.close()


from adhesive.storage.ensure_folder import ensure_folder



from typing import Optional, Dict, Any

from adhesive.graph.BaseTask import BaseTask
from adhesive.steps.Execution import Execution
from adhesive.steps.ExecutionData import ExecutionData
from adhesive.workspace.Workspace import Workspace


class ExecutionToken:
    """
    A context passed to an execution of a task. It holds the information
    about:
    - data that's attached to this token,
    - workspace where files can be created. This depends on the actual runtime
      (ie linux, windows, docker)
    - loop information (when in a loop).

    A workflow context it's an execution token that's being passed around.
    """
    def __init__(self,
                 task: 'BaseTask',
                 execution: Execution,
                 data: Optional[Dict],
                 workspace: Workspace) -> None:
        self.task = task
        self.data = ExecutionData(data)
        self.execution = execution
        self.workspace = workspace

        self.loop: Optional[WorkflowLoop] = None

        self.update_title()

    def update_title(self) -> None:
        # FIXME: this breaks the encapsulation of the data
        try:
            self.task_name = self.task.name.format(**{
                "context": self,
                "execution": self.execution,
                "data": self.data,
                "loop": self.loop,
            })
        except Exception as e:
            self.task_name = self.task.name

    def clone(self, task: 'BaseTask') -> 'ExecutionToken':
        result = ExecutionToken(
            task,
            self.execution,
            self.data.as_dict(),
            self.workspace,  # FIXME: this should be a clone
        )

        return result

    def as_mapping(self) -> Dict[str, Any]:
        """
        This mapping is for evaluating routing conditions.
        :return:
        """
        return {
            "task": self.task,
            "execution": self.execution,
            "data": self.data,
            "loop": self.loop,
            "task_name": self.task_name,
        }


from adhesive.steps.WorkflowLoop import WorkflowLoop



from typing import Optional, List, Dict, Any, cast, Tuple, Union, Iterable

import npyscreen as npyscreen

from adhesive.model.ActiveEvent import ActiveEvent
from adhesive.model.UiBuilderApi import UiBuilderApi
from adhesive.model.UserTaskProvider import UserTaskProvider
from adhesive.model.WorkflowExecutor import WorkflowExecutor
from adhesive.steps.AdhesiveUserTask import AdhesiveUserTask
from adhesive.steps.ExecutionData import ExecutionData


class UIBuilder(UiBuilderApi):
    def __init__(self,
                 event: ActiveEvent):
        self.context = event.context
        self.ui_controls: Dict[str, Any] = dict()

        self.labels: Dict[str, List[str]] = dict()
        self.values: Dict[str, List[str]] = dict()

        self.ncurses_calls = []

    @property
    def data(self) -> ExecutionData:
        result_dict = dict()

        for name, ui_control in self.ui_controls.items():
            if isinstance(self.ui_controls[name], npyscreen.TitleMultiSelect):
                result = {self.values[name][it] for it in ui_control.value}
                result_dict[name] = result
                continue

            if isinstance(self.ui_controls[name], npyscreen.TitleCombo):
                result = self.values[name][ui_control.value] if ui_control.value >= 0 else None
                result_dict[name] = result
                continue

            if isinstance(self.ui_controls[name], npyscreen.TitleSelectOne):
                result = self.values[name][ui_control.value[0]] if len(ui_control.value) > 0 else None
                result_dict[name] = result
                continue

            result_dict[name] = ui_control.value

        return ExecutionData(result_dict)

    def add_input_text(self,
                       name: str,
                       title: Optional[str] = None,
                       value: str = '') -> None:
        if not name:
            raise Exception("You need to pass a name for the input.")

        if not title:
            title = name

        def ncurses_input_text_call():
            self.ui_controls[name] = self.form.add_widget(
                npyscreen.TitleText,
                name=title,
                value=value)

        self.ncurses_calls.append(ncurses_input_text_call)

    def add_input_password(self,
                       name: str,
                       title: Optional[str] = None,
                       value: str = '') -> None:
        if not name:
            raise Exception("You need to pass a name for the input.")

        if not title:
            title = name

        def ncurses_input_password_call():
            self.ui_controls[name] = self.form.add_widget(
                npyscreen.TitlePassword,
                name=title,
                value=value)

        self.ncurses_calls.append(ncurses_input_password_call)

    def add_combobox(
            self,
            name: str,
            title: Optional[str] = None,
            value: Optional[str]=None,
            values: Optional[Iterable[Union[Tuple[str, str], str]]]=None) -> None:
        if not name:
            raise Exception("You need to pass a name for the input.")

        if not title:
            title = name

        self.values[name] = UIBuilder._get_values(values)
        self.labels[name] = UIBuilder._get_labels(values)

        _value = self.values[name].index(UIBuilder._get_value(value)) if value else -1

        def ncurses_add_combobox_call():
            self.ui_controls[name] = self.form.add_widget(
                npyscreen.TitleCombo,
                name=title,
                value=_value,
                values=self.labels[name])

        self.ncurses_calls.append(ncurses_add_combobox_call)

    def add_checkbox_group(
            self,
            name: str,
            title: Optional[str]=None,
            value: Optional[Iterable[str]]=None,
            values: Optional[Iterable[Union[Tuple[str, str], str]]]=None) -> None:
        if not name:
            raise Exception("You need to pass a name for the input.")

        if not values:
            raise Exception("You need to pass in some values to display in the group.")

        if not title:
            title = name

        if value is None:
            value = []

        title_overflows = 0 if len(title) < 14 else 1

        self.values[name] = UIBuilder._get_values(values)
        self.labels[name] = UIBuilder._get_labels(values)

        def ncurses_add_checbox_group_call():
            self.ui_controls[name] = self.form.add_widget(
                npyscreen.TitleMultiSelect,
                name=title,
                value=[self.values[name].index(UIBuilder._get_value(v)) for v in value],
                max_height=max(len(values) + 1, 2) + title_overflows,
                scroll_exit=True,
                values=self.labels[name])

        self.ncurses_calls.append(ncurses_add_checbox_group_call)

    def add_radio_group(self,
                        name: str,
                        title: Optional[str]=None,
                        value: Optional[str]=None,
                        values: Optional[Iterable[Union[Tuple[str, str], str]]]=None) -> None:
        if not name:
            raise Exception("You need to pass a name for the input.")

        if not title:
            title = name

        title_overflows = 0 if len(title) < 14 else 1

        self.values[name] = UIBuilder._get_values(values)
        self.labels[name] = UIBuilder._get_labels(values)

        _value = self.values[name].index(UIBuilder._get_value(value)) if value is not None else -1

        def ncurses_add_radio_group_call():
            self.ui_controls[name] = self.form.add_widget(
                npyscreen.TitleSelectOne,
                name=title,
                max_height=max(len(values), 2) + title_overflows,
                scroll_exit=True,
                value=_value,
                values=self.labels[name])

        self.ncurses_calls.append(ncurses_add_radio_group_call)

    def add_default_button(self,
                           name: str,
                           title: Optional[str]=None) -> None:
        self.context.data[name] = name

    @staticmethod
    def _get_values(values) -> List[str]:
        result = []

        for value in values:
            if isinstance(value, str):
                result.append(value)
                continue

            result.append(value[0])

        return result

    @staticmethod
    def _get_value(value) -> str:
        if isinstance(value, str):
            return value

        return value[0]

    @staticmethod
    def _get_labels(values) -> List[str]:
        result = []

        for value in values:
            if isinstance(value, str):
                result.append(value)
                continue

            result.append(value[1])

        return result


class ConsoleUserTaskProvider(UserTaskProvider):
    def __init__(self):
        super(ConsoleUserTaskProvider, self).__init__()

    def register_event(self,
                       executor: WorkflowExecutor,
                       event: ActiveEvent) -> None:

        ui = UIBuilder(event)

        adhesive_task = cast(AdhesiveUserTask, executor.tasks_impl[event.task.id])
        context = adhesive_task.invoke_user_task(event, ui)

        # redirecting logs, and initializing ncurses is prolly a bad idea
        # the code that generates the UI shouldn't be run in ncurses
        def run_on_curses(x):
            try:
                # build the UI components on ncurses:
                ui.form = npyscreen.Form(name=event.task.name)

                for ncurses_call in ui.ncurses_calls:
                    ncurses_call()

                # call the actual UI
                ui.form.edit()

                # put the values in the data.
                context.data = ExecutionData.merge(context.data, ui.data)

                event.future.set_result(context)
            except Exception as e:
                event.future.set_exception(e)

        npyscreen.wrapper_basic(run_on_curses)
